# Generated by Django 4.1.2 on 2022-11-10 01:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_q", "0014_schedule_cluster"),
    ]

    operations = [
        migrations.AlterField(
            model_name="schedule",
            name="schedule_type",
            field=models.CharField(
                choices=[
                    ("O", "Once"),
                    ("I", "Minutes"),
                    ("H", "Hourly"),
                    ("D", "Daily"),
                    ("W", "Weekly"),
                    ("BW", "Biweekly"),
                    ("M", "Monthly"),
                    ("BM", "Bimonthly"),
                    ("Q", "Quarterly"),
                    ("Y", "Yearly"),
                    ("C", "Cron"),
                ],
                default="O",
                max_length=2,
                verbose_name="Schedule Type",
            ),
        ),
    ]
